///////////////////////////////////////////////////////////////////////////
//
// Copyright (c) 2004 Intel Corporation 
// All rights reserved. 
//
// Redistribution and use in source and binary forms, with or without 
// modification, are permitted provided that the following conditions are met: 
//
// * Redistributions of source code must retain the above copyright notice, 
// this list of conditions and the following disclaimer. 
// * Redistributions in binary form must reproduce the above copyright notice, 
// this list of conditions and the following disclaimer in the documentation 
// and/or other materials provided with the distribution. 
// * Neither name of Intel Corporation nor the names of its contributors 
// may be used to endorse or promote products derived from this software 
// without specific prior written permission.
// 
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS 
// "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT 
// LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR 
// A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL INTEL OR 
// CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, 
// EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, 
// PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR 
// PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY 
// OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
// NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS 
// SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
//
///////////////////////////////////////////////////////////////////////////

#include "stdafx.h"
#include "ipp.h"

#include "affine.h"
#include "Image.h"

Affine::Affine() : isSrcROI_(0), isDstROI_(0), interp_(IPPI_INTER_LINEAR)
{
	Reset();
}

Affine::~Affine() {}

void Affine::SetInterpType(int type)
{
	if (type == NN) interp_ = IPPI_INTER_NN;
	else if (type == LINEAR) interp_ = IPPI_INTER_LINEAR;
	else if (type == CUBIC) interp_ = IPPI_INTER_CUBIC;
}

int Affine::GetInterpType() const
{
	if (interp_ == IPPI_INTER_NN) return NN;
	else if (interp_ == IPPI_INTER_LINEAR) return LINEAR;
	else if (interp_ == IPPI_INTER_CUBIC) return CUBIC;
	else return 0;
}

void Affine::SetSmoothMode(int smoothFlag)
{
	if (smoothFlag) interp_ |= IPPI_SMOOTH_EDGE;
	else interp_ &= ~IPPI_SMOOTH_EDGE;
}

void Affine::SetSrcROI(IppiRect srcROI)
{
	srcROI_.x = srcROI.x;
	srcROI_.y = srcROI.y;
	srcROI_.width = srcROI.width;
	srcROI_.height = srcROI.height;

	isSrcROI_ = 1;
}

//add destWindow? (differentiate between ROI and shift)
//maybe just "set shift" that overrides default shift
void Affine::SetDestROI(IppiRect dstROI)
{
	dstROI_.x = dstROI.x;
	dstROI_.y = dstROI.y;
	dstROI_.width = dstROI.width;
	dstROI_.height = dstROI.height;
	
	isDstROI_ = 1;
}

IppiSize Affine::GetDestSize_(IppiRect srcROI) const
{
	double bound[2][2];
	double xShift = 0.0, yShift = 0.0;
	IppiSize dstSize = {0,0};

//	if (isCenter_)
//		ippiGetAffineShift(xCenter_, yCenter_, angle_, &xShift, &yShift);
	ippiGetAffineBound(srcROI, bound, pCoeffs_);

	dstSize.width = (int)(bound[1][0]+0.5) - (int)(bound[0][0]-0.5) + 1.0;
	dstSize.height = (int)(bound[1][1]+0.5) - (int)(bound[0][1]-0.5) + 1.0;
	if (isDstROI_)
	{
		if (dstSize.width > (dstROI_.width + dstROI_.x))
			dstSize.width = dstROI_.width +dstROI_.x;
		if (dstSize.height > (dstROI_.height + dstROI_.y))
			dstSize.height = dstROI_.height + dstROI_.y;
	}
	return dstSize;
}

IppiSize Affine::GetDestSize() const
{
	IppiSize dstSize = {0,0};
	if (isSrcROI_) dstSize = GetDestSize_(srcROI_);
	return dstSize;
}

IppiSize Affine::GetDestSize(IppiSize srcSize) const
{
	IppiSize dstSize;
	IppiRect srcROI;
	if (isSrcROI_)
	{
		dstSize = GetDestSize_(srcROI_);
	}
	else
	{
		srcROI.x = srcROI.y = 0;
		srcROI.width = srcSize.width;
		srcROI.height = srcSize.height;
		dstSize = GetDestSize_(srcROI);
	}

	return dstSize;
}

void Affine::Reset()
{
	double pCoeffs[2][3] = { { 1.0, 0.0, 0.0 }, {0.0, 1.0, 0.0 } };
	SetAffine(pCoeffs);
}

void Affine::AddRotate(double degrees)
{
	double s, c;

	degrees = degrees * IPP_PI / 180.0;
	ippsSinCos_64f_A50(&degrees, &s, &c, 1);

	double pCoeffs[2][3] = { { c, s, 0.0}, {-s, c, 0.0} };
	AddAffine(pCoeffs);
}

void Affine::AddShear(double xShear, double yShear)
{
	double pCoeffs[2][3] = { { 1.0, xShear, 0.0}, {yShear, 1.0, 0.0} };
	AddAffine(pCoeffs);
}
void Affine::AddResize(double xFactor, double yFactor)
{
	double pCoeffs[2][3] = { { xFactor, 0.0, 0.0}, {0.0, yFactor, 0.0} };
	AddAffine(pCoeffs);
}

void Affine::AddAffine(const double pCoeffs[2][3])
{
	double s00, s01, s02, s10, s11, s12;
	double d00, d01, d02, d10, d11, d12;

	s00 = pCoeffs[0][0]; s10 = pCoeffs[1][0];
	s01 = pCoeffs[0][1]; s11 = pCoeffs[1][1];
	s02 = pCoeffs[0][2]; s12 = pCoeffs[1][2];

	d00 = pCoeffs_[0][0]; d10 = pCoeffs_[1][0];
	d01 = pCoeffs_[0][1]; d11 = pCoeffs_[1][1];
	d02 = pCoeffs_[0][2]; d12 = pCoeffs_[1][2];

	//								[x]
	//								[y]
	//								[1]
	//					[00 01 02]
	//					[10 11 12]
	//					[-  -   1]  
	//		[00 01 02]
	//		[10 11 12]   
	//		[-  -   1]

	pCoeffs_[0][0] = s00 * d00 + s01 * d10;
	pCoeffs_[1][0] = s10 * d00 + s11 * d10;

	pCoeffs_[0][1] = s00 * d01 + s01 * d11;
	pCoeffs_[1][1] = s10 * d01 + s11 * d11;

	pCoeffs_[0][2] = s00 * d02 + s01 * d12 + s02;
	pCoeffs_[1][2] = s10 * d02 + s11 * d12 + s12;
}

void Affine::SetAffine(const double pCoeffs[2][3])
{
	pCoeffs_[0][0] = pCoeffs[0][0];
	pCoeffs_[0][1] = pCoeffs[0][1];
	pCoeffs_[0][2] = pCoeffs[0][2];
	pCoeffs_[1][0] = pCoeffs[1][0];
	pCoeffs_[1][1] = pCoeffs[1][1];
	pCoeffs_[1][2] = pCoeffs[1][2];
}

int Affine::Go(const Image8u*pSrc, Image8u* pDst)
{
	IppStatus st;

	double xShift=0.0, yShift=0.0;
	double bound[2][2];

	int channelCount = pSrc->GetChannelCount();
	if (channelCount != pDst->GetChannelCount()) return -1;

	if (!isDstROI_) {
		dstROI_.x = dstROI_.y = 0;
		dstROI_.width = pDst->GetSize().width;
		dstROI_.height = pDst->GetSize().height;
	}
	if (!isSrcROI_) {
		srcROI_.x = srcROI_.y = 0;
		srcROI_.width = pSrc->GetSize().width;
		srcROI_.height = pSrc->GetSize().height;
	}

	ippiGetAffineBound(srcROI_, bound, pCoeffs_);
	pCoeffs_[0][2] -= bound[0][0];
	pCoeffs_[1][2] -= bound[0][1];

	if (channelCount == 1)
		st = ippiWarpAffine_8u_C1R(pSrc->GetConstData(),
			pSrc->GetSize(), pSrc->GetStride(), srcROI_,
			pDst->GetData(), pDst->GetStride(), dstROI_,
			pCoeffs_, interp_);
	else if (channelCount == 3)
		st = ippiWarpAffine_8u_C3R(pSrc->GetConstData(),
			pSrc->GetSize(), pSrc->GetStride(), srcROI_,
			pDst->GetData(), pDst->GetStride(), dstROI_,
			pCoeffs_, interp_);
	else if (channelCount == 4)
		st = ippiWarpAffine_8u_C4R(pSrc->GetConstData(),
			pSrc->GetSize(), pSrc->GetStride(), srcROI_,
			pDst->GetData(), pDst->GetStride(), dstROI_,
			pCoeffs_, interp_);
	else return -1;

	pCoeffs_[0][2] += bound[0][0];
	pCoeffs_[1][2] += bound[0][1];

	return ((int)st);
}
